#include "repetition_testing/reptester.h"
#include "profiler/timer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/time.h>

void handle_alloc(reptester *tester, alloc_type type) {
  switch (type) {
  case ALLOC_TYPE_WITH_MALLOC:
    if (!(tester->params.buffer)) {
      tester->params.buffer = (char *)mmap(
          NULL, tester->params.read_size + 1, PROT_READ | PROT_WRITE,
          MAP_ANON | MAP_SHARED | MAP_NORESERVE, -1, 0);
      memset(tester->params.buffer, 0, tester->params.read_size + 1);
    }

    break;
  default:
    break;
  }
}

void handle_free(reptester *tester, alloc_type type) {
  switch (type) {
  case ALLOC_TYPE_WITH_MALLOC:
    if (tester->params.buffer) {
      munmap(tester->params.buffer, tester->params.read_size + 1);
      tester->params.buffer = NULL;
    }

    break;
  default:
    break;
  }
}

void run_func_test(reptester *tester, reptest_func func, const char *func_name,
                   alloc_type type) {
  tester->test_start_time = read_cpu_timer();
  tester->test_time_secs = 0.0;
  tester->current_run = 1;
  tester->tstats = {
      UINT64_MAX, // min_time
      0,          // max_time
      0,          // avg_time
      0,          // total_time
  };
  tester->mstats = {
      UINT64_MAX, // min_faults
      0,          // max_faults
      0,          // avg_faults
      0,          // total_bytes
      0,          // total_faults
  };
  tester->results = {};

  char *buffer = NULL;

  if (type == ALLOC_TYPE_WITH_MALLOC) {
    buffer = tester->params.buffer;
    tester->params.buffer =
        (char *)mmap(NULL, tester->params.read_size + 1, PROT_READ | PROT_WRITE,
                     MAP_ANON | MAP_SHARED | MAP_NORESERVE, -1, 0);
    memset(tester->params.buffer, 0, tester->params.read_size + 1);
  }

  while (tester->test_time_secs <= tester->wait_time_secs) {
    func(tester, type);

    if (tester->results.bytes_read <
        tester->params.read_size * tester->params.read_count) {
      printf("Failed to read the entire file (Total size: %lu, Bytes read: "
             "%lu)\n",
             tester->params.read_size, tester->results.bytes_read);

      return;
    }

    tester->tstats.total_time += tester->results.read_time;
    tester->mstats.total_bytes += tester->results.bytes_read;
    tester->mstats.total_faults += tester->results.page_faults;

    if (tester->results.read_time > tester->tstats.max_time) {
      tester->tstats.max_time = tester->results.read_time;
      tester->mstats.max_faults = tester->results.page_faults;
    } else if (tester->results.read_time < tester->tstats.min_time) {
      tester->test_start_time = read_cpu_timer();
      tester->tstats.min_time = tester->results.read_time;
      tester->mstats.min_faults = tester->results.page_faults;
    }

    if (tester->results.page_faults > tester->mstats.max_faults) {
      tester->mstats.max_faults = tester->results.page_faults;
    } else if (tester->results.page_faults < tester->mstats.min_faults) {
      tester->mstats.min_faults = tester->results.page_faults;
    }

    tester->test_time_secs = time_in_seconds(
        read_cpu_timer() - tester->test_start_time, tester->cpu_freq);

    ++(tester->current_run);
  }

  if (type == ALLOC_TYPE_WITH_MALLOC) {
    munmap(tester->params.buffer, tester->params.read_size + 1);
    tester->params.buffer = buffer;
  }

  print_results(tester, func_name);
}

u64 page_fault_count() {
  rusage usage;

  getrusage(RUSAGE_SELF, &usage);

  return usage.ru_minflt + usage.ru_majflt;
}

void print_results(reptester *tester, const char *name) {
  f64 kb = 1024.0;
  f64 gb = kb * kb * kb;

  f64 size_in_kb =
      (f64)(tester->params.read_size * tester->params.read_count) / kb;
  f64 size_in_gb =
      (f64)(tester->params.read_size * tester->params.read_count) / gb;

  u64 run_count = tester->current_run - 1;

  tester->tstats.avg_time = tester->tstats.total_time / run_count;
  tester->mstats.avg_faults = tester->mstats.total_faults / run_count;

  printf("\n%s: %lu runs\n", name, run_count);

  printf("MIN: %lu (%fGB/s)", tester->tstats.min_time,
         size_in_gb /
             time_in_seconds(tester->tstats.min_time, tester->cpu_freq));
  if (tester->mstats.min_faults > 0) {
    printf(", FAULTS: %lu (%fK/fault)\n", tester->mstats.min_faults,
           size_in_kb / tester->mstats.min_faults);
  } else {
    printf("\n");
  }

  printf("MAX: %lu (%fGB/s)", tester->tstats.max_time,
         size_in_gb /
             time_in_seconds(tester->tstats.max_time, tester->cpu_freq));
  if (tester->mstats.max_faults > 0) {
    printf(", FAULTS: %lu (%fK/fault)\n", tester->mstats.max_faults,
           size_in_kb / tester->mstats.max_faults);
  } else {
    printf("\n");
  }

  printf("AVG: %lu (%fGB/s)", tester->tstats.avg_time,
         size_in_gb /
             time_in_seconds(tester->tstats.avg_time, tester->cpu_freq));
  if (tester->mstats.avg_faults > 0) {
    printf(", FAULTS: %lu (%fK/fault)\n", tester->mstats.avg_faults,
           tester->mstats.total_bytes / kb / tester->mstats.avg_faults);
  } else {
    printf("\n");
  }
}
